package org.fortiss.tooling.kernel.ui.internal;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.IEclipseResourcePostLoadProvider;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.constraints.ConstrainedWithChecksum;
import org.fortiss.tooling.kernel.model.constraints.ConstraintsPackage;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraint;
import org.fortiss.tooling.kernel.model.constraints.IConstraintVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedVerificationStatus;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.ui.extension.IConstraintVerifierUI;
import org.fortiss.tooling.kernel.ui.service.IConstraintVerificationUIService;
import org.fortiss.tooling.kernel.ui.util.ConstraintsUIUtils;
import org.fortiss.tooling.kernel.utils.ConstraintsUtils;
import org.fortiss.tooling.kernel.utils.EcoreUtils;
import org.fortiss.tooling.kernel.utils.LoggingUtils;

/**
 * Implementation of {@link IConstraintVerificationService}.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: CB3FC2AF2D45F99291553F698F3957B9
 */
public final class ConstraintVerificationUIService extends
		EObjectAwareServiceBase<IConstraintVerifierUI<IConstraint>> implements
		IConstraintVerificationUIService, IEclipseResourcePostLoadProvider {

	/**
	 * @param constraint
	 * @return the first registered verifier which provides a GUI for the verifier of
	 *         <code>constraint</code>
	 */
	private IConstraintVerifierUI<IConstraint> getFirstVerifier(IConstraint constraint) {
		List<IConstraintVerifierUI<IConstraint>> list =
				getRegisteredHandlers(constraint.getClass());
		if(list == null || list.isEmpty()) {
			LoggingUtils.error(ToolingKernelActivator.getDefault(), String.format(
					"Cannot find verifierUI for constraint %s", constraint.getClass()));
			return null;
		}
		// get(0) because we know the list is not empty and because we want the first element
		return list.get(0);
	}

	/** {@inheritDoc} */
	@Override
	public void openStatus(IConstraintVerificationStatus status) {
		IConstraintVerifierUI<IConstraint> verifier = getFirstVerifier(status.getConstraint());
		verifier.openStatus(status);
		ConstraintsUIUtils.triggerMarkersRefresh(status.getConstraint());
	}

	/** {@inheritDoc} */
	@Override
	public boolean canOpen(IConstraintVerificationStatus status) {
		if(status != null) {
			IConstraintVerifierUI<IConstraint> verifier = getFirstVerifier(status.getConstraint());
			return verifier.canOpen(status);
		}
		return false;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * Installs all the required notifiers.
	 */
	@Override
	public void postModelLoadRun(ITopLevelElement element) {
		TreeIterator<EObject> tree = element.getRootModelElement().eAllContents();
		while(tree.hasNext()) {
			EObject next = tree.next();
			if(next instanceof IConstraint) {
				installNotifiers((IConstraint)next);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void installNotifiers(IConstraint constraint) {
		new ConstrainedAdapters(constraint);
	}

	/** Adapter for changes on constrained elements. */
	private static class ConstrainedAdapters {

		/** Map containing the adapter of each constrained element. */
		Map<IConstrained, Adapter> constrainedAdapterMap;

		/** Constructor. */
		public ConstrainedAdapters(IConstraint constraint) {
			constrainedAdapterMap = new HashMap<IConstrained, Adapter>();
			for(ConstrainedWithChecksum cwc : constraint.getConstrainedsWithChecksum()) {
				IConstrained constrained = cwc.getConstrained();
				Adapter adapter = new EContentAdapter() {
					@Override
					public void notifyChanged(Notification notification) {
						super.notifyChanged(notification);
						if(isTouch(notification)) {
							return;
						}
						if(constraintRemoved(notification, constrained)) {
							return;
						}
						constraintOutdated(constraint, notification);
					}
				};
				constrained.eAdapters().add(adapter);
				constrainedAdapterMap.put(constrained, adapter);
			}
		}

		/**
		 * @param constraint
		 * @param notification
		 *            Checks if <code>notification</code> denotes a change entailing that
		 *            <code>constraint</code> is outdated. If so updates the status to outdated or
		 *            triggers an automatic check if the <code>autocheck</code> field of the
		 *            constraint is <code>true</code>.
		 * @return <code>true</code> iff <code>notification</code> indeed entailed that
		 *         <code>constraint</code> is outdated.
		 */
		private boolean constraintOutdated(IConstraint constraint, Notification notification) {
			if(skipStatusChangeNotification(constraint, notification)) {
				return false;
			}
			if(!IConstraintVerificationService.INSTANCE.isUpToDate(constraint)) {
				if(constraint.isAutoCheck()) {
					ITopLevelElement modelContext =
							IPersistencyService.INSTANCE.getTopLevelElementFor(constraint);
					modelContext.runAsCommand(new Runnable() {
						@Override
						public void run() {
							IConstraintVerificationService.INSTANCE.verify(constraint);
						}
					});
				} else if(!(constraint.getVerificationStatus() instanceof OutdatedVerificationStatus)) {
					ConstraintsUtils.createOutdatedVerificationStatus(constraint);
				}
				// Update GUI markers
				ConstraintsUIUtils.triggerMarkersRefresh(constraint);
				return true;
			}
			return false;
		}

		/**
		 * @param constraint
		 * @param notification
		 * @return true iff <code>notification</code> denotes a change of the
		 *         <code>constraint</code>'s status.
		 */
		private boolean skipStatusChangeNotification(IConstraint constraint,
				Notification notification) {
			String n = ConstraintsPackage.eINSTANCE.getIConstraint_VerificationStatus().getName();
			if(notification.getNotifier() != null && notification.getNotifier().equals(constraint)) {
				if(notification.getFeature() instanceof EReference) {
					return(n.equals(((ENamedElement)notification.getFeature()).getName()));
				}
			}
			return false;
		}

		/**
		 * @param notification
		 * @param constrained
		 * @return true if the notification is a consequence of a constraint of
		 *         <code>constrained</code> having been removed.
		 */
		private boolean constraintRemoved(Notification notification, IConstrained constrained) {
			if(notification.getEventType() != Notification.REMOVE) {
				return false;
			}
			if(!notification.getNotifier().equals(constrained)) {
				return false;
			}
			if(notification.getFeature() instanceof EReference) {
				String name = ((ENamedElement)notification.getFeature()).getName();
				if(name.equals(ConstraintsPackage.eINSTANCE.getIConstrained_Constraints().getName())) {
					constrained.eAdapters().remove(constrainedAdapterMap.get(constrained));
					constrainedAdapterMap.remove(constrained);
					return true;
				}
			}
			return true;
		}
	}

	/**
	 * @param notification
	 * @return true if <code>notification</code> is not relevant and should be skipped
	 */
	private static boolean isTouch(Notification notification) {
		// Notifications which do not notify about a model change.
		if(notification.isTouch()) {
			return true;
		}
		// Notifications initiated by the the AF3 kernel which are only GUI related.
		if(notification.getEventType() == EcoreUtils.KERNEL_EMF_EVENT_TYPE_COUNT) {
			return true;
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return "org.fortiss.tooling.kernel.ui.constraintVerifierUI";
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return "constraintVerifierUI";
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return "constraintVerifierUI";
	}

	/** {@inheritDoc} */
	@Override
	protected String getClassAttribute() {
		return "constraint";
	}

	/** {@inheritDoc} */
	@Override
	protected String getClassConfigurationElement() {
		return "constraint";
	}
}
